import logging
from itertools import islice
from .utils import wait_for_redis_ready

_log = logging.getLogger(__name__)


def _get_config():
  import phovea_server.config
  return phovea_server.config.view('phovea_data_redis.mapping')


def create_db():
  import redis

  c = _get_config()

  # print c.hostname, c.port, c.db
  return redis.Redis(host=c.hostname, port=c.port, db=c.db, charset='utf-8', decode_responses=True, **c.extras)


class RedisMappingTable(object):
  def __init__(self, from_idtype, to_idtype):
    self.from_idtype = from_idtype
    self.to_idtype = to_idtype

  def __call__(self, ids):
    db = create_db()

    def map_impl(id):
      key = '{}2{}.{}'.format(self.from_idtype, self.to_idtype, id)
      v = ''
      if db.get(key) is not None:
        v = db.get(key)
      return v.split(';')

    return [map_impl(id) for id in ids]

  def search(self, query, max_results=None):
    """
    searches for matches in the names of the given idtype
    :param query:
    :param max_results
    :return:
    """
    db = create_db()
    query = ''.join(('[' + lower + upper + ']' for lower, upper in zip(query.upper(), query.lower())))
    prefix = '{}2{}.'.format(self.from_idtype, self.to_idtype)
    match = '{}*{}*'.format(prefix, query)
    keys = [k for k in islice(db.scan_iter(match=match), max_results)]
    values = db.mget(keys)
    return [dict(match=key[len(prefix):], to=value) for key, value in zip(keys, values)]


class CachedRedisMappingTable(object):
  def __init__(self, from_idtype, to_idtype, db, all_keys):
    self.from_idtype = from_idtype
    self.to_idtype = to_idtype
    self._cache = self._load_cache(db, all_keys)

  def _load_cache(self, db, all_keys):
    import fnmatch
    prefix = '{}2{}.'.format(self.from_idtype, self.to_idtype)
    match = prefix + '*'
    keys = [k for k in fnmatch.filter(all_keys, match)]
    values = db.mget(keys)
    return {key[len(prefix):]: value for key, value in zip(keys, values)}

  def __call__(self, ids):
    def map_impl(id):
      v = self._cache.get(id, '')
      return v.split(';')

    return [map_impl(id.decode('utf-8')) for id in ids]

  def search(self, query, max_results=None):
    """
    searches for matches in the names of the given idtype
    :param query:
    :param max_results
    :return:
    """
    query = query.lower()
    return [dict(match=key, to=self._cache[key]) for key in
            islice((k for k in list(self._cache.keys()) if query in k.lower()), max_results)]


def _discover_mappings():
  db = create_db()
  wait_for_redis_ready(db)
  mappings = db.get('mappings')
  _log.info('found %s', mappings)
  if not mappings:
    return
  mappings = [r for r in mappings.split(';') if r.strip()]
  cached = _get_config().bulk
  if cached:
    keys = list(db.keys())
  else:
    keys = None
  for key in mappings:
    parts = key.split('2')
    _log.info('loading redis mapping table from %s to %s', parts[0], parts[1])
    from_idtype = parts[0]
    to_idtype = parts[1]
    yield (CachedRedisMappingTable(from_idtype, to_idtype, db, keys) if cached else RedisMappingTable(from_idtype, to_idtype))


class RedisMappingProvider(object):
  def __init__(self):
    self._mappings = list(_discover_mappings())

  def __iter__(self):
    return iter(((f.from_idtype, f.to_idtype, f) for f in self._mappings))


def create():
  return RedisMappingProvider()



import logging
from .utils import wait_for_redis_ready

_log = logging.getLogger(__name__)


class RedisCache(object):
  """
  assigns ids to object using a redis database
  """

  def __init__(self):
    import redis
    import phovea_server.config

    c = phovea_server.config.view('phovea_data_redis.cache')

    # print c.hostname, c.port, c.db
    self._db = redis.Redis(host=c.hostname, port=c.port, db=c.db, charset='utf-8', decode_responses=True, **c.extras)
    wait_for_redis_ready(self._db)
    self._default_timeout = c.timeout

  def get(self, key):
    return self._db.get(key)

  def set(self, key, value, timeout=None):
    self._db.set(key, value)
    if timeout is None:
      timeout = self._default_timeout
    if timeout > 0:
      self._db.expire(key, timeout)


def create():
  _log.info('create redis cache')
  return RedisCache()



import logging

_log = logging.getLogger(__name__)


def wait_for_redis_ready(db, timeout=None):
  import time
  import redis

  _log.info('check if redis is ready')
  start_time = time.time()
  while timeout is None or time.time() - start_time < timeout:
    try:
      db.dbsize()
    except redis.exceptions.BusyLoadingError:
      _log.info('stall till redis is ready')
      time.sleep(0.5)
    else:
        break



import logging
from codecs import open

_log = logging.getLogger(__name__)


def _get_db(key):
  import redis
  import phovea_server.config

  c = phovea_server.config.view(key)

  return redis.Redis(host=c.hostname, port=c.port, db=c.db, charset='utf-8', decode_responses=True)


def _get_assigner_db():
  return _get_db('phovea_data_redis.assigner')


def _get_mapping_db():
  return _get_db('phovea_data_redis.mapping')


def remove_all_ids():
  db = _get_assigner_db()
  db.flushdb()


def remove_all_mappings():
  db = _get_mapping_db()
  db.flushdb()


def load_ids_from_file(idtype, file_name, set_max=True):
  db = _get_assigner_db()
  _log.info('loading ids for %s from %s', idtype, file_name)
  with open(file_name, 'r', encoding='utf-8') as f:
    max_uid = None
    for line in f:
      [id, uid] = line.split('\t')
      uid = int(uid)
      if max_uid is None or uid > max_uid:
        max_uid = uid
      key = '{}2id.{}'.format(idtype, id)
      db.set(key, uid)
      key = 'id2{}.{}'.format(idtype, uid)
      db.set(key, id)
  if set_max:
    db.set(idtype, max_uid + 1)


def load_mapping_from_file(from_idtype, to_idtype, file_name):
  db = _get_mapping_db()
  _log.info('loading mapping from %s -> %s from %s', from_idtype, to_idtype, file_name)

  db.append('mappings', '{}2{};'.format(from_idtype, to_idtype))
  with open(file_name, 'r', encoding='utf-8') as f:
    for line in f:
      s = line.split('\t')
      key = '{}2{}.{}'.format(from_idtype, to_idtype, s[0].strip())
      db.set(key, ';'.join(s[1:]).strip())



import logging
from itertools import islice

_log = logging.getLogger(__name__)


def ascii(s):
  return s


class RedisIDAssigner(object):
  """
  assigns ids to object using a redis database
  """

  def __init__(self):
    import redis
    import phovea_server.config
    from .utils import wait_for_redis_ready

    c = phovea_server.config.view('phovea_data_redis.assigner')

    # print c.hostname, c.port, c.db
    self._db = redis.Redis(host=c.hostname, port=c.port, db=c.db, charset='utf-8', decode_responses=True, **c.extras)
    wait_for_redis_ready(self._db)

    from cachelib import SimpleCache
    self._cache = SimpleCache(threshold=c.cache)

  @staticmethod
  def to_forward_key(idtype, identifier):
    return ascii(idtype + '2id.' + str(identifier))

  @staticmethod
  def from_forward_key(idtype, forward_key):
    prefix = idtype + '2id.'
    return forward_key[len(prefix):]

  @staticmethod
  def to_backward_key(idtype, id):
    return ascii('id2' + idtype + '.' + str(id))

  def _get_entry(self, key):
    r = self._cache.get(key)
    if r is not None:  # cached
      return r
    r = self._db.get(key)  # lookup in db
    if r is not None:  # cache if found
      self._cache.set(key, r)
    return r

  def _get_entries(self, keys):
    result = []
    not_cached = []
    not_cached_indices = []
    for i, k in enumerate(keys):
      cached = self._cache.get(k)
      result.append(cached)
      if cached is None:
        not_cached.append(k)
        not_cached_indices.append(i)

    if len(not_cached) > 0:
      values = self._db.mget(not_cached)
      for k, i, v in zip(not_cached, not_cached_indices, values):
        if v:
          self._cache.set(k, v)
          result[i] = v
    return result

  def unmap(self, uids, idtype):
    return self._get_entries((self.to_backward_key(idtype, id) for id in uids))

  def load(self, idtype, mapping):
    """
    resets and loads the given mapping
    :param idtype:
    :param mapping: array of tuples (id, uid)
    :return:
    """

    # assuming incremental ids
    if idtype in self._db:
      _log.info('clearing %s', idtype)
      self._cache.clear()
      forward_keys = self._db.keys(idtype + '2id.*')
      _log.debug('found %d forward keys', len(forward_keys))
      self._db.delete(forward_keys)
      backward_keys = self._db.keys('id2' + idtype + '.*')
      _log.debug('found %d backwards keys', len(backward_keys))
      self._db.delete(backward_keys)
      _log.debug('deleted %d keys', len(backward_keys) + len(forward_keys))
      del self._db[idtype]

    if len(mapping) == 0:
      return  # no mappings to load

    max_uid = None
    pipe = self._db.pipeline()

    _log.info('%s loading %s keys', idtype, len(mapping))
    for id, uid in mapping:
      key = self.to_forward_key(idtype, id)
      max_uid = uid if max_uid is None else max(uid, max_uid)
      pipe.set(key, uid)
      pipe.set(self.to_backward_key(idtype, uid), str(id))

    pipe.set(idtype, max_uid)
    _log.debug('flushing')

    pipe.execute()
    _log.debug('flushed')

  def __call__(self, ids, idtype):
    """
     return the integer index ids for the given ids in the given idtype
    """
    idtype = ascii(idtype)
    if not isinstance(ids, list):
      ids = list(ids)  # maybe a generator

    before = int(self._db.get(idtype) if idtype in self._db else self._db.decr(idtype))  # initialize with -1
    entries = self._get_entries((self.to_forward_key(idtype, id) for id in ids))

    def create_entry(i, entry):
      if entry is None:
        id = ids[i]
        key = self.to_forward_key(idtype, id)
        entry = self._db.incr(idtype)
        self._db.set(key, entry)
        self._cache.set(key, entry)
        back_key = self.to_backward_key(idtype, entry)
        self._db.set(back_key, str(id))
        self._cache.set(back_key, str(id))
      return entry

    r = [int(create_entry(i, entry)) for i, entry in enumerate(entries)]

    after = int(self._db.get(idtype))
    if before != after:
      _log.debug('create %s %d!=%d', idtype, before, after)

    return r

  def search(self, idtype, query, max_results=None):
    """
    searches for matches in the names of the given idtype
    :param idtype:
    :param query:
    :param max_results
    :return:
    """
    query = ''.join(('[' + lower + upper + ']' for lower, upper in zip(query.upper(), query.lower())))
    match = self.to_forward_key(idtype, '*' + query + '*')
    keys = [k for k in islice(self._db.scan_iter(match=match), max_results)]
    ids = self._get_entries(keys)
    return [dict(id=int(id), name=self.from_forward_key(idtype, key)) for key, id in zip(keys, ids)]


def create():
  _log.info('create redis assigner')
  return RedisIDAssigner()
